from django import forms
from django.contrib import admin
from django.utils.text import capfirst
from django.utils.translation import ugettext_lazy as _

from .api import SOURCES
from .models import Condition, Generator


class ConditionInline(admin.TabularInline):
    model = Condition
    extra = 0


@admin.register(Generator)
class GeneratorAdmin(admin.ModelAdmin):
    inlines = [ConditionInline]
    list_display = ["group", "source", "get_conditions_display"]
    ordering = ["source", "group"]

    def get_queryset(self, request):
        return self.model.objects.prefetch_related("conditions")

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        if db_field.name == "source":
            kwargs["widget"] = forms.Select(
                choices=list(
                    sorted(
                        (
                            (key, capfirst(source.get("verbose_name", key)))
                            for key, source in SOURCES.items()
                        ),
                        key=lambda row: row[1],
                    )
                )
            )
        return super().formfield_for_dbfield(db_field, request=request, **kwargs)

    def get_conditions_display(self, instance):
        return " && ".join(str(condition) for condition in instance.conditions.all())

    get_conditions_display.short_description = _("conditions")



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SparkGeneratorsConfig(AppConfig):
    name = "spark.spark_generators"
    verbose_name = _("Spark generators")



import operator

from django.db import models
from django.utils.translation import ugettext_lazy as _


class Generator(models.Model):
    group = models.CharField(_("group"), max_length=50)
    source = models.CharField(_("source"), max_length=50)

    class Meta:
        verbose_name = _("generator")
        verbose_name_plural = _("generators")

    def __str__(self):
        return self.group

    def as_generator(self):
        return {
            "group": self.group,
            "source": self.source,
            "conditions": [
                condition.as_condition() for condition in self.conditions.all()
            ],
        }


class Condition(models.Model):
    OPERATORS = [
        (">", operator.gt),
        ("<", operator.lt),
        (">=", operator.ge),
        ("<=", operator.le),
        ("=", operator.eq),
        ("!=", operator.ne),
    ]

    generator = models.ForeignKey(
        Generator,
        on_delete=models.CASCADE,
        related_name="conditions",
        verbose_name=_("generator"),
    )
    variable = models.CharField(_("variable"), max_length=50)
    operator = models.CharField(
        _("operator"), max_length=10, choices=[(d, d) for d, op in OPERATORS]
    )
    value = models.IntegerField(_("value"))

    class Meta:
        ordering = ["pk"]
        verbose_name = _("condition")
        verbose_name_plural = _("conditions")

    def __str__(self):
        return "{o.variable} {o.operator} {o.value}".format(o=self)

    def as_condition(self):
        op = dict(self.OPERATORS)[self.operator]
        return {"variable": self.variable, "test": lambda v: op(v, self.value)}



# Generated by Django 2.0 on 2017-12-19 14:26

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Event",
            fields=[
                (
                    "key",
                    models.CharField(
                        max_length=100,
                        primary_key=True,
                        serialize=False,
                        verbose_name="key",
                    ),
                ),
                (
                    "group",
                    models.CharField(
                        db_index=True, max_length=50, verbose_name="group"
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="created at"),
                ),
            ],
            options={"verbose_name": "events", "ordering": ["-created_at"]},
        )
    ]



from django.core.management.base import BaseCommand

from spark import api


class Command(BaseCommand):
    help = "Sends all registered models through all event sources"

    def handle(self, **options):
        for model, sources in api.MODEL_SOURCES.items():
            for instance in model.objects.all():  # FIXME allow restrictions
                for source in sources:
                    api.process_events(source(instance))



import re

from django.db.models import signals


MODEL_SOURCES = {}
HANDLERS = []


def process_events(iterable):
    from .models import Event

    for e in iterable:
        if Event.objects.create_if_new(e):
            for group, handler in HANDLERS:
                if re.search(group, e["group"]):
                    handler(e)


def register_model_event_source(*, sender, source):
    MODEL_SOURCES.setdefault(sender, []).append(source)

    def on_post_save(sender, instance, **kwargs):
        process_events(source(instance))

    signals.post_save.connect(on_post_save, sender=sender, weak=False)


def register_group_handler(*, group, handler):
    HANDLERS.append((group, handler))



# Generated by Django 2.0 on 2017-12-19 14:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("spark", "0001_initial")]

    operations = [
        migrations.AlterModelOptions(
            name="event",
            options={
                "ordering": ["-created_at"],
                "verbose_name": "event",
                "verbose_name_plural": "events",
            },
        ),
        migrations.AddField(
            model_name="event",
            name="context",
            field=models.TextField(blank=True, verbose_name="context"),
        ),
    ]



# Generated by Django 2.1.2 on 2018-10-26 07:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Condition",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("variable", models.CharField(max_length=50, verbose_name="variable")),
                (
                    "operator",
                    models.CharField(
                        choices=[
                            (">", ">"),
                            ("<", "<"),
                            (">=", ">="),
                            ("<=", "<="),
                            ("=", "="),
                            ("!=", "!="),
                        ],
                        max_length=10,
                        verbose_name="operator",
                    ),
                ),
                ("value", models.IntegerField(verbose_name="value")),
            ],
            options={
                "verbose_name": "condition",
                "verbose_name_plural": "conditions",
                "ordering": ["pk"],
            },
        ),
        migrations.CreateModel(
            name="Generator",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("group", models.CharField(max_length=50, verbose_name="group")),
                ("source", models.CharField(max_length=50, verbose_name="source")),
            ],
            options={"verbose_name": "generator", "verbose_name_plural": "generators"},
        ),
        migrations.AddField(
            model_name="condition",
            name="generator",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="conditions",
                to="spark_generators.Generator",
                verbose_name="generator",
            ),
        ),
    ]



from django.contrib import admin

from spark.models import Event


@admin.register(Event)
class EventAdmin(admin.ModelAdmin):
    list_display = ("key", "group", "created_at", "context")
    list_filter = ("group", "created_at")



from django.db import IntegrityError, models, transaction
from django.utils.translation import ugettext_lazy as _


class EventQuerySet(models.QuerySet):
    def create_if_new(self, event):
        context = dict(event)
        try:
            with transaction.atomic():
                return self.create(
                    key=context.pop("key"),
                    group=context.pop("group"),
                    context=repr(context),
                )
        except IntegrityError:
            return None


class Event(models.Model):
    key = models.CharField(_("key"), max_length=100, primary_key=True)
    group = models.CharField(_("group"), max_length=50, db_index=True)
    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    context = models.TextField(_("context"), blank=True)

    objects = EventQuerySet.as_manager()

    class Meta:
        ordering = ["-created_at"]
        verbose_name = _("event")
        verbose_name_plural = _("events")

    def __str__(self):
        return self.key



#!/usr/bin/env python

import os
from setuptools import setup, find_packages


def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


setup(
    name='django-spark',
    version=__import__('spark').__version__,
    description='Event sourcing and handling',
    long_description=read('README.rst'),
    author='Matthias Kestenholz',
    author_email='mk@feinheit.ch',
    url='http://github.com/matthiask/django-spark/',
    license='BSD License',
    platforms=['OS Independent'],
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    zip_safe=False,
)



from spark.models import Event
from spark.spark_generators.models import Generator


SOURCES = {}


def pure_function_memoizer():
    MEMO = {}

    def call(fn, *args):
        key = (fn, *args)
        if key not in MEMO:
            MEMO[key] = fn(*args)
        return MEMO[key]

    return call


def events_from_generators(queryset=None):
    queryset = Generator.objects.all() if queryset is None else queryset
    memoizer = pure_function_memoizer()

    for generator in queryset.prefetch_related("conditions"):
        g = generator.as_generator()

        source = SOURCES[g["source"]]
        candidates = memoizer(source["candidates"])
        for candidate in candidates:
            variables = source["variables"](candidate)
            for condition in g["conditions"]:
                if condition["variable"] not in variables:
                    break
                if not condition["test"](variables[condition["variable"]]):
                    break
            else:
                e = source["event"](
                    instance=candidate, generator=g, variables=variables
                )
                if Event.objects.create_if_new(e):
                    yield e
